--[[
Name: Bits-1.0
Revision: $Revision: 25921 $
Author(s): Karl The Pagan (karlthepagan@gmail.com)
Website: http://www.wowace.com/
Documentation: 
SVN: http://svn.wowace.com/wowace/trunk/Bits/
Description: Provides a set-based front end to the Lua bitlib library
Dependencies: AceLibrary, (optional) AceComm-2.0
]]

--[[
"Bits-1.0" Provides a set-based front end to the Lua bitlib library
Copyright (C) 2006 Karl Pauls

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

The text of the GNU Lesser General Public License is available at
the following URL: http://www.gnu.org/copyleft/lesser.txt

To receive a copy of the GNU Lesser General Public License write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301  USA
]]

local MAJOR_VERSION = "Bits-1.0"
local MINOR_VERSION = "$Revision: 25921 $"

if AceLibrary and not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local AceOO = AceLibrary and AceLibrary:HasInstance("AceOO-2.0") and AceLibrary("AceOO-2.0")

local function activate(self, oldLib, oldDeactivate)
	if oldDeactivate then oldDeactivate(oldLib) end
end

-- importing bitlib
local lshift,band,bor,bxor,bnot = bit.lshift,bit.band,bit.bor,bit.bxor,bit.bnot

-- precalc powers (faster hash creation)
local BIT_LIMIT = 53
local BIT_VALUES = {};
for i=1,BIT_LIMIT do
    BIT_VALUES[i] = lshift(1,i-1)
end

-- OPTIONAL AceOO.Mixin
local BitSet = AceOO and (AceOO.Mixin {
    "Size", -- the set size
    "Clone", -- copy a bit set (optionally into toTarget)
    "Add", -- add additional set elements
    "Alias", -- assigns a new key aliased to the combination (bit.bor) of the given keys
	"Get", -- get hash values for keys
    "GetHash", -- get hash
    "AddToHash", -- add item to hash
    "RemoveFromHash", -- remove item from hash
    "MaskHash", -- mask hash against these bits, ignoring the others - first return is only these bits, second is the ones removed
    "ToggleInHash", -- toggles bits in the hash
    "HashEquals", -- test a hash for equality with the set of keys
    "HashContains", -- test if a hash contains any of a set of keys
    "ReverseMap", -- retrieve a reverse map of the bit set elements which maps values to the keys
    "DeHash", -- break down a list of hashes into lists of keys
    "And", -- logical and of key values
    "Or", -- logical or of key values
    "Xor", -- logical exclusive or of key values "only works with aliases" (unless you understand the math)
}) or {}

local Bits = {}
function Bits:new(...)
    local set = {_n = 0}
    
    if AceOO then
        BitSet:embed(set)
    else
        BitSet:Clone(set)
    end
    
    set:Add(...)
    
    return set
end

function BitSet:Size()
    return self._n
end

function BitSet:Clone(toTarget)
    if not toTarget or type(toTarget) ~= "table" then
        toTarget = {}
    end
    
    for k,v in pairs(self) do
        if not toTarget[k] then
            toTarget[k] = v
        end
    end
end

function BitSet:Add(...)
	for x = 1,select('#',...) do
        local v = select(x,...)
        
        if self[v] then
            error("cannot overwrite Bits key '"..v.."'")
        end
        
        -- see README for bitlib -- 8-byte IEEE float provides 53 bits
        if self._n > BIT_LIMIT-1 then
            error("Set can only accept "..BIT_LIMIT.." elements")
        end
        
        self._n = self._n + 1
        self[v] = BIT_VALUES[self._n]
	end
end

function BitSet:Get(...)
    -- thanks Pastamancer and Iriel for optimizations & suggestions
    local len = select("#", ...)
    
    if len == 0 then
        return
    end
    
    return (self[(select(1,...))] or 0), self:Get(select(2, ...))
end

function BitSet:GetHash(...)
	return self:Or(...)
end

function BitSet:AddToHash(initial,...)
    assert(initial)
	return bor(initial,self:Get(...))
end

function BitSet:RemoveFromHash(initial,...)
    assert(initial)
	return band(initial,
		bnot(bor(0,self:Get(...))))
end

function BitSet:MaskHash(initial,...)
    assert(initial)
    local mask = bor(0,self:Get(...))
    return band(initial,mask),band(initial,bnot(mask))
end

function BitSet:ToggleInHash(initial,...)
    assert(initial)
    return bxor(initial,
        bor(0,self:Get(...)))
end

function BitSet:HashEquals(hash,...)
    return hash == bor(0,self:Get(...))
end

function BitSet:HashContains(hash,...)
    return 0 ~= band(hash,bor(0,self:Get(...)))
end

function BitSet:ReverseMap()
    local reverse = {}
    
    for i = 1,self._n do
        reverse[BIT_VALUES[i]] = true
    end
    
    for k,v in pairs(self) do
        if reverse[v] == true then
            reverse[v] = k
        end
    end
    
    return reverse
end

local function deHashStep(lim,reverse,...)
    -- thanks Pastamancer and Iriel for optimizations & suggestions
    local v = select(1,...)
    
    if not v then
        return
    end
    
    local keys = {}
    local i = 1
    
    for i = 1,lim do
        local x = band(v,BIT_VALUES[i])
        if x ~= 0 then
            table.insert(keys,reverse[x])
        end
    end
        
    return keys, deHashStep(lim,reverse,select(2,...))
end

function BitSet:DeHash(...)
    local reverse = self:ReverseMap()
    
    return deHashStep(self._n,reverse,...)
end

function BitSet:And(...)
    if select('#', ...) < 1 then return 0 end
	return band(self:Get(...))
end

function BitSet:Or(...)
	return bor(0,self:Get(...))
end

function BitSet:Xor(...)
    return bxor(0,self:Get(...))
end

function BitSet:Alias(key,...)
    if self[key] then
        error("cannot overwrite Bits key '"..key.."'")
    end
    
    self[key] = self:Or(...)
end


if AceLibrary then AceLibrary:Register(Bits, MAJOR_VERSION, MINOR_VERSION)
else getfenv(0).Bits = Bits end
